# -*- coding: utf-8 -*-
"""
Tests for Dark.py
"""


import os
import shutil
import tempfile

import numpy

from micado.config.Environment import Env
from micado.hardware.ShotNoise import ShotNoise
from micado.main.ObservingBlock import ObservingBlock
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.DarkCurrent import DarkCurrent
from micado.dataitems.Dark import Dark
from micado.dataitems.DarkExtension import DarkExtension
from micado.templates.MCD_img_cal_dark import MCD_img_cal_dark
from micado.util.esokeywords import verify_fitsheaders


def test_dark():
    """
    Test creation of a Dark.
    """
    Env['use_python_logging'] = True
    # Set the real dark current and create a DetectorArray.
    dark_real = 10000.
    mydetector = Detector()
    mydetectorlist = [mydetector]
    mydarkcurrent = DarkCurrent(value=dark_real)
    myshotnoise = ShotNoise(state=True)
    mydetectorarray = DetectorArray(
        name='testdetector',
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
        shot_noise=myshotnoise,
    )

    # A template
    dark_template = MCD_img_cal_dark(det__dit=1, det__ndit=1)

    # An observingblock
    dark_ob = ObservingBlock(
        name="MyDarkTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[dark_template],
    )

    # DarkExtensions
    myextensions = [
        DarkExtension(detector=det, BITPIX=-64, NAXIS=2)
        for det in mydetectorlist
    ]

    dark1 = Dark(
        detector=mydetectorarray,
        template=dark_template,
        observing_block=dark_ob,
        extensions=myextensions,
    )
    dark1.make_simulation()

    dark1.save_fits()
    fn = dark1.filename
    verify_fitsheaders(fn)

    dark2 = Dark(filename=fn)

    assert numpy.all(dark2.extensions[0].image == dark1.extensions[0].image)
    assert dark1.extensions[0].NAXIS1 == dark2.extensions[0].NAXIS1
    assert dark1.extensions[0].NAXIS2 == dark2.extensions[0].NAXIS2
    assert dark1.detector == dark2.detector
    assert dark1.template == dark2.template
    assert dark1.observing_block == dark2.observing_block
    ext1 = dark1.extensions[0]
    ext2 = dark2.extensions[0]
    assert ext1.detector == ext2.detector
    assert ext1 == ext2
    assert dark2 == dark1

    fn3 = tempfile.NamedTemporaryFile(suffix='.fits', delete=False).name
    shutil.copy(fn, fn3)

    dark3 = Dark(filename=fn3)
    dark3.unique_filename()

    hash1 = dark1.filename.split("_")[-1]
    hash3 = dark3.filename.split("_")[-1]
    assert hash1 == hash3, "Bad file hashes"

    os.remove(dark1.filename)
    os.remove(dark1.filename_short)
    os.remove(dark3.filename)
    os.remove(dark3.filename_short)
    os.remove(fn3)


def test_dark_multi():
    """
    Test creation of a Dark.
    """
    Env['use_python_logging'] = True
    # Set the real dark current and create a DetectorArray.
    dark_real = 10000.
    mydarkcurrent = DarkCurrent(value=dark_real)

    mydetector1 = Detector(id="det1", x_cen=0)
    mydetector2 = Detector(id="det2", x_cen=400)
    mydetectorlist = [mydetector1, mydetector2]
    myshotnoise = ShotNoise(state=True)

    mydetectorarray = DetectorArray(
        name='testdetector',
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
        shot_noise=myshotnoise,
    )

    # A template
    dark_template = MCD_img_cal_dark(det__dit=1)

    # An observingblock
    dark_ob = ObservingBlock(
        name="MyDarkTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[dark_template],
    )

    mydarkextensions = [
        DarkExtension(detector=mydetector, BITPIX=-64, NAXIS=2)
        for mydetector in mydetectorlist
    ]

    for myext in mydarkextensions:
        myext.image = numpy.ones(
            (myext.NAXIS1, myext.NAXIS2),
            dtype=numpy.float64,
        )

    dark1 = Dark(
        detector=mydetectorarray,
        template=dark_template,
        observing_block=dark_ob,
        extensions=mydarkextensions,
    )

    from pprint import pprint
    pprint(dark1.fits_keywords)
    dark1.save_fits()
    fn = dark1.filename

    dark2 = Dark(filename=fn)

    assert dark2.detector.dark_current.value == dark1.detector.dark_current.value
    assert dark2.detector == dark1.detector
    assert dark2.extensions[0] == dark1.extensions[0]
    assert dark2.extensions[1] == dark1.extensions[1]
    assert dark2 == dark1

    os.remove(dark1.filename)
    os.remove(dark1.filename_short)
