# -*- coding: utf-8 -*-
"""
Tests for Science.py

TODO: Should these tests be automated?
      Most of this could be copied verbatim from Dark_test.py
      Maybe easy once backward chaining is properly working.
"""


import os
import shutil
import tempfile

import numpy

from micado.config.Environment import Env
from micado.hardware.ShotNoise import ShotNoise
from micado.main.ObservingBlock import ObservingBlock
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.DarkCurrent import DarkCurrent
from micado.dataitems.Science import Science
from micado.dataitems.ScienceExtension import ScienceExtension
from micado.templates.MCD_img_sci_object import MCD_img_sci_object
from micado.util.esokeywords import verify_fitsheaders


def test_science():
    """
    Test creation of a Raw Science image.
    """
    Env['use_python_logging'] = True
    # Set the dark current and create a DetectorArray.
    dark_current = 100.
    mydetector = Detector()
    mydetectorlist = [mydetector]
    mydarkcurrent = DarkCurrent(value=dark_current)
    myshotnoise = ShotNoise(state=True)

    mydetectorarray = DetectorArray(
        name='testdetector',
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
        shot_noise=myshotnoise,
    )

    # A template
    science_template = MCD_img_sci_object(det__dit=1)

    # An observingblock
    science_ob = ObservingBlock(
        name="MyScienceTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[science_template],
    )

    # Extensions
    myextensions = [
        ScienceExtension(detector=det, BITPIX=-64, NAXIS=2)
        for det in mydetectorlist
    ]

    science1 = Science(
        detector=mydetectorarray,
        template=science_template,
        observing_block=science_ob,
        extensions=myextensions,
    )
    science1.make_simulation()

    science1.save_fits()
    fn = science1.filename

    science2 = Science(filename=fn)

    assert numpy.all(science2.extensions[0].image == science1.extensions[0].image)
    assert science1.extensions[0].NAXIS1 == science2.extensions[0].NAXIS1
    assert science1.extensions[0].NAXIS2 == science2.extensions[0].NAXIS2
    assert science1.detector == science2.detector
    assert science1.template == science2.template
    assert science1.observing_block == science2.observing_block
    ext1 = science1.extensions[0]
    ext2 = science2.extensions[0]
    assert ext1.detector == ext2.detector
    assert ext1 == ext2
    assert science2 == science1

    fn3 = tempfile.NamedTemporaryFile(suffix='.fits', delete=False).name
    shutil.copy(fn, fn3)

    science3 = Science(filename=fn3)
    science3.unique_filename()

    hash1 = science1.filename.split("_")[-1]
    hash3 = science3.filename.split("_")[-1]
    assert hash1 == hash3, "Bad file hashes"

    verify_fitsheaders(fn)

    os.remove(science1.filename)
    os.remove(science1.filename_short)
    os.remove(science3.filename)
    os.remove(science3.filename_short)
    os.remove(fn3)


def test_science_multi():
    """
    Test creation of a multi-extension raw image.
    """
    Env['use_python_logging'] = True
    # Set the dark current and create a DetectorArray.
    dark_current = 100.
    mydarkcurrent = DarkCurrent(value=dark_current)

    mydetector1 = Detector(id="det1", x_cen=0)
    mydetector2 = Detector(id="det2", x_cen=400)
    mydetectorlist = [mydetector1, mydetector2]
    myshotnoise = ShotNoise(state=True)

    mydetectorarray = DetectorArray(
        name='testdetector',
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
        shot_noise=myshotnoise,
    )

    # A template
    science_template = MCD_img_sci_object(det__dit=1)

    # An observingblock
    science_ob = ObservingBlock(
        name="MyScienceTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[science_template],
    )

    myscienceextensions = [
        ScienceExtension(detector=mydetector, BITPIX=-64, NAXIS=2)
        for mydetector in mydetectorlist
    ]

    for myext in myscienceextensions:
        myext.image = numpy.ones(
            (myext.NAXIS1, myext.NAXIS2),
            dtype=numpy.float64,
        )

    science1 = Science(
        detector=mydetectorarray,
        template=science_template,
        observing_block=science_ob,
        extensions=myscienceextensions,
    )

    from pprint import pprint
    pprint(science1.fits_keywords)
    science1.save_fits()
    fn = science1.filename

    science2 = Science(filename=fn)

    assert science2.detector.dark_current.value == science1.detector.dark_current.value
    assert science2.detector == science1.detector
    assert science2.extensions[0] == science1.extensions[0]
    assert science2.extensions[1] == science1.extensions[1]
    assert science2 == science1

    os.remove(science1.filename)
    os.remove(science1.filename_short)
