# -*- coding: utf-8 -*-
"""
Generate .ditaa figures from DataItems.
"""

from collections import OrderedDict

from pathlib import Path

from micado.toolbox.ditaacreator import (
    SimpleStepInput, SimpleStep, SimpleRecipe,
    DataRaw, DataExternal, DataFinal, DataIntermediate,
    RecipeStep as RecipeStepDitaa,
    RecipeFull, Process, RecipeBox,
    Template as TemplateDitaa,
    SimpleStepParams, ParamsBox)

from micado.main.RecipeStep import RecipeStep
from micado.main.Img import Img
from micado.main.DataItem import DataItem
from micado.main.Raw import Raw
from micado.main.Ext import Ext
from micado.main.Final import Final
from micado.dataitems.MasterDark import MasterDark
from micado.dataitems.MasterFlat import MasterFlat
from micado.dataitems.IllumCorr import IllumCorr
from micado.main.StaticTable import StaticTable
from micado.dataitems.Detrend import Detrend

PATH_DOCS_GENERATED = Path(__file__).resolve().parent.parent.parent.parent / "docs" / "source" / "generated"


def get_ditaaclass(attrtype):
    ditaaclass_from_wiseclass = OrderedDict([
        (Raw, DataRaw),
        (Final, DataFinal),
        (Ext, DataExternal),
        (Img, DataIntermediate),
        (DataItem, DataIntermediate),
    ])
    for clsi, clsd in ditaaclass_from_wiseclass.items():
        if issubclass(attrtype, clsi):
            clsn = clsd
            break
    else:
        clsn = DataIntermediate
    return clsn


def get_step(mo: RecipeStep):
    inputs = [SimpleStepInput(
        cls=get_ditaaclass(attrtype),
        text=attrtype.name_eso(),
    ) for attrtype in mo.get_other_inputdata()]
    params = [pn.upper() for pn in mo.get_parameter_names()]

    return SimpleStep(
        process=mo.name_eso(),
        inputs=inputs,
        params=SimpleStepParams(text=params) if len(params) else None,
        raw=mo.get_raw().name_eso() if mo.get_raw() is not None else None,
        previous=get_step(mo.get_previous_step()) if mo.get_previous_step() is not None else None,
    )


def ditaa_from_wise(mo: type):
    """
    Create a ditaa file from an
    """
    if not issubclass(mo, (Img, StaticTable)):
        print("ditaa_from_wise can only be ran on processed Img or StaticTable dataitems")
        return ""

    recipe = getattr(mo, 'recipe').prop_type.name_eso()
    rr = mo.get_raw_class()
    temp = rr.template.prop_type
    template = temp.__name__

    raw = rr.name_eso()
    steps = []
    mmo = mo.recipe.prop_type
    rs = mmo.process.prop_type
    while rs is not None:
        step = get_step(rs)
        steps.append(step)
        rs = rs.get_previous_step()

    clsp = get_ditaaclass(mo)

    sr = SimpleRecipe(
        template=template,
        raw=raw,
        recipe=recipe,
        steps=steps[::-1],
        product=mo.name_eso(),
    )

    asteps = [
        RecipeStepDitaa(
            process=Process(step.process),
            inputs=[n.cls(n.text) for n in step.inputs],
            params=ParamsBox(text=step.params.text) if step.params is not None else None,
            raw=DataRaw(step.raw) if step.raw is not None else None,
            previous=Process(step.previous.process) if step.previous is not None else None,
        ) for step in sr.steps
    ]

    rf = RecipeFull(
        template=TemplateDitaa(sr.template),
        raw=DataRaw(sr.raw),
        steps=asteps,
        recipe=RecipeBox(text=sr.recipe),
        product=clsp(sr.product)
    )
    return rf


def camel_from_snake(value):
    """
    Convert apply_master_dark back to ApplyMasterDark
    """
    words = [word.capitalize() for word in value.split('_')]
    return "".join(words)


def main():
    for cls in [MasterDark, MasterFlat, IllumCorr, Detrend]:
        fn = PATH_DOCS_GENERATED / f"{cls.__name__}.ditaa"
        rf = ditaa_from_wise(cls)
        s = str(rf)
        print(fn)
        with open(fn, "w") as f:
            f.write(s)

        for rs in rf.steps:
            s = "\n".join(rs.lines_selfcontained)
            # TODO: Really, really refactor this Ditaa creation stuff..
            #       Put some of the logic with the classes themselves or so.
            fn = PATH_DOCS_GENERATED / f"{camel_from_snake(rs.process.text)}.ditaa"
            print(fn)
            with open(fn, "w") as f:
                f.write(s)


if __name__ == '__main__':
    main()
