# -*- coding: utf-8 -*-
"""
Create Python modules from a VODML/XML file.

TODO: Cleanup this script
TODO: Use type annotations
TODO: Use units
TODO: Add to aweimports.py
"""

from collections import OrderedDict
from dataclasses import dataclass, field
from lxml import etree
from pathlib import Path
from typing import List, Optional

from micado.util.esokeywords import esokeywords


@dataclass
class Attribute:
    """
    An attribute of an ObjectType.
    """
    name: str
    description: str = ""
    unit: str = "None"
    stype: str = ""
    default: str = ""
    vodmlid: str = ""
    esokeyword: str = ""


@dataclass
class FitsKeyword:
    """A FITS Keyword"""
    keyword: str
    vodmlid: str
    wiseid: str


def default_objecttype():
    """
    Default ObjectType
    """
    return OBJECTTYPES['MObject']


@dataclass
class ObjectType:
    """
    An ObjectType.
    """
    name: str
    extends: "Optional[ObjectType]" = field(default_factory=default_objecttype)
    description: str = ""
    imports: List[str] = field(default_factory=list)
    attributes: List[Attribute] = field(default_factory=list)
    abstract: bool = False

    def issubtype(self, other):
        """
        Check whether this is a subtype.
        """
        if other.name == self.name:
            return True
        elif isinstance(self.extends, ObjectType):
            return self.extends.issubtype(other)
        else:
            return False

    def ishardware(self):
        """Is Hardware?"""
        return self.issubtype(OBJECTTYPES['Hardware'])

    def isdataitem(self):
        """Is Dataitem?"""
        return self.issubtype(OBJECTTYPES['DataItem'])

    def __str__(self):
        attrs = [
            '    {} = persistent("{} [{}]", {}, {})'.format(
                aa.name, aa.description, aa.unit, aa.stype, aa.default
            ) for aa in self.attributes
        ]
        desc_indent = "\n".join(f"    {l}" for l in self.description.split("\n"))

        simports = "\n".join(self.imports)
        sattrs = "\n\n".join(attrs)
        sfile = "\n".join([
            f'# -*- coding: utf-8 -*-',
            f'"""',
            f'{self.description}',
            f'',
            f'This file is automatically generated. Do not modify!',
            f'"""',
            f'',
            f'{simports}',
            f'',
            f'',
            f'class {self.name}({self.extends.name}):',
            f'    """',
            f'{desc_indent}',
            f'    """',
            f'',
            f'{sattrs}',
            f'',
        ])
        return sfile

    def getattr(self, attrname):
        """
        Get an attribute.
        TODO: actually use __getattr__? or make attributes an OrderedDict?
        """
        attrd = {a.name: a for a in self.attributes}
        return attrd.get(attrname, None)


@dataclass
class Package:
    """
    A Package.
    """
    name: str = ""
    path: Path = field(default_factory=Path)
    object_types: List[ObjectType] = field(default_factory=list)


def import_from_ref(ref):
    """
    micado:hardware.DetectorArray
    from micado.hardware.DetectorArray import DetectorArray
    """
    refq, refc = ref.split(".")
    refm = ref.replace(":", ".")

    imp = "from {} import {}".format(refm, refc)
    return imp


def parse_objecttype(ot):
    """
    Parse one ObjectType
    """
    name_ot = ot.find('name').text

    description1 = ot.find('description').text
    # TODO: Perhaps raise if there is no description?
    description2 = description1 if description1 else "None"
    description = "\n".join(l.strip() for l in description2.split("\n"))

    element_extends = ot.find('extends')
    extends = element_extends.find('vodml-ref').text if element_extends is not None else 'micado:main.MObject'
    extends_single = extends.split(".")[-1]
    attributes = ot.findall('attribute') + ot.findall('reference') + ot.findall('composition')
    is_abstract = ot.get("abstract") == 'true'

    toimport = list()
    toimport.append(import_from_ref(extends) + ", persistent")

    lattrs = [Attribute(
        name=f'dummy_{name_ot.lower()}',
        description="dummy",
        stype="int",
        default="42",
    )]
    for elattr in attributes:
        name_attr = elattr.find('name').text
        desc_attr = elattr.find('description').text
        type_attr = elattr.find('datatype').find('vodml-ref').text
        el_multi = elattr.find('multiplicity')
        if el_multi is not None:
            is_list = el_multi.find('maxOccurs').text == '-1'
        else:
            is_list = False

        if type_attr.startswith('ivoa'):
            # TODO: datetime
            type_singlet = {
                'ivoa:integer': int,
                'ivoa:real': float,
                'ivoa:IntegerQuantity': int,
                'ivoa:RealQuantity': float,
                'ivoa:string': str,
                'ivoa:boolean': bool,
            }.get(type_attr)
            type_default = {
                int: 0,
                float: 0.0,
                str: '""',
                bool: False,
            }.get(type_singlet)
            type_single = type_singlet.__name__
            if is_list:
                type_default = [type_default]

        else:
            simport = import_from_ref(type_attr)
            if simport not in toimport:
                toimport.append(simport)
            type_single = type_attr.split(".")[-1]
            type_default = None
            if is_list:
                type_default = []

        vodmlid = elattr.find('vodml-id').text
        if desc_attr in esokeywords:
            esokeyword = desc_attr
            esoattr = esokeywords[esokeyword]
            the_default = esoattr.get('default', type_default)
            if the_default in (None, 'None'):
                the_default = type_default
            desc_attr2 = esoattr.get('description')
            unit_attr = esoattr.get('unit')
        else:
            the_default = type_default
            desc_attr2 = desc_attr
            unit_attr = "None"
            esokeyword = ""

        lattrs.append(Attribute(
            name=name_attr,
            description=desc_attr2,
            stype=type_single,
            default=the_default,
            vodmlid=vodmlid,
            unit=unit_attr,
            esokeyword=esokeyword,
        ))

    assert extends_single in OBJECTTYPES, f"{extends_single} of {name_ot} not defined"
    assert name_ot not in OBJECTTYPES, f"{name_ot} already defined"

    otfull = ObjectType(
        name=name_ot,
        extends=OBJECTTYPES[extends_single],
        description=description,
        imports=toimport,
        attributes=lattrs,
        abstract=is_abstract,
    )
    OBJECTTYPES[otfull.name] = otfull
    print(f"{otfull.name} {otfull.extends.name if otfull.extends is not None else 'None'}")
    return otfull


def parse_package(package, path_gen):
    """
    Parse one package
    """
    name_package = package.find('name').text
    path_package = path_gen / name_package

    object_types = package.findall('objectType')

    ots = []
    for ot in object_types:
        ot = parse_objecttype(ot)
        ots.append(ot)

    package = Package(
        name=name_package,
        path=path_package,
        object_types=ots
    )
    return package


def main(verify=False):
    """
    Convert VODML/XML to Python classes.
    """
    p = Path(__file__).resolve().parent.parent.parent / "vodml"
    fn = p / "micado.vo-dml.xml"
    path_gen = Path(__file__).parent.parent
    assert fn.exists(), fn
    assert path_gen.exists()

    tree = etree.parse(str(fn))

    packages = tree.findall('package')

    for xpackage in packages:
        package = parse_package(xpackage, path_gen)
        print("PACKAGE", package.name)
        # 'main' is done manually.
        if package.name in ["hardware", "dataitems", "recipes", "templates"]:
            for ot in package.object_types:
                # Abstract classes are done manually.
                if not ot.abstract:
                    path_ot = package.path / f"{ot.name}.py"
                    if verify:
                        with path_ot.open() as f_ot:
                            ot_expected = f_ot.read()
                            assert ot_expected == str(ot)
                    else:
                        with path_ot.open("w") as f_ot:
                            f_ot.write(str(ot))

    for ot in OBJECTTYPES.values():
        print(f"{ot.name} {ot.ishardware()} {ot.isdataitem()}")


OBJECTTYPES = OrderedDict([('MObject', ObjectType(name='MObject', extends=None))])

if __name__ == '__main__':
    main()
