# -*- coding: utf-8 -*-
"""
Create rst files from Python.
"""
import os
from collections import OrderedDict

import yaml
from micado.dataitems.IllumCorr import IllumCorr
from micado.dataitems.MasterFlat import MasterFlat
from micado.main.RecipeStep import RecipeStep
from pathlib import Path

from micado.dataitems.MasterDark import MasterDark
from micado.dataitems.Detrend import Detrend

from micado.toolbox.ditaa_from_wise import PATH_DOCS_GENERATED
from micado.util.esokeywords import esokeywords

PATH_TEXT_YAML = Path(__file__).resolve().parent.parent.parent / "text"


def table_from_dataitem(dataitem):
    """
    Gerenare table from dataitem.
    """
    raw = dataitem.get_raw_class()
    template = raw.template.prop_type
    qc = dataitem.get_qc_class()

    asbad = ['filename', 'globalname', 'object_id']

    # Get QC parameters
    if qc:
        asqcopt = [
            a for a in qc.get_persistent_properties()
            if a not in asbad and not a.startswith('dummy')
        ]
        sasqc = [f"QC.{a.upper()}" for a in asqcopt]
        sqc = ", ".join(sasqc)
    else:
        sqc = "None"

    # Get template parameters
    if template:
        astemopt = [
            a for a in template.get_persistent_properties()
            if a not in asbad and not a.startswith('dummy')
        ]
        as_ok = [a for a in astemopt if not a.startswith('dpr')]
        sastemplate = [a.upper().replace("__", ".") for a in as_ok]
        stemplatepars = ", ".join(sastemplate)
    else:
        stemplatepars = "None"

    # Get input data
    step: RecipeStep = dataitem.recipe.prop_type.process.prop_type
    inputdata = [raw]
    while step:
        inputdata += step.get_other_inputdata()
        step = step.get_previous_step()

    ssinput = [f"{dat.name_eso()}_" for dat in inputdata]
    sinput = ", ".join(ssinput)

    # TODO: Replace with dataclasses?
    fn_raw = PATH_TEXT_YAML / "raw" / f"{raw.__name__.lower()}.yaml"
    fn_pro = PATH_TEXT_YAML / "pro" / f"{dataitem.__name__.lower()}.yaml"
    if os.path.exists(fn_raw):
        dic_raw = yaml.load(open(fn_raw).read())
    else:
        dic_raw = {
            "responsible": " ",
            "phase": " ",
            "micadosetup": " ",
            "calunitsetup": " ",
            "frequency": " ",
            "duration": " ",
            "prerequisites": " ",
        }
    if os.path.exists(fn_pro):
        dic_pro = yaml.load(open(fn_pro).read())
    else:
        dic_pro = {
            "name": " ",
            "objective": " ",
            "frequency": " ",
            "procedure": " ",
            "math": " ",
            "remarks": " ",
            "prosepurpose": " ",
            "proseworkflow": " ",
        }

    dic1 = OrderedDict([
        ("Name", dic_pro["name"]),
        ("ReqID", "TODO: IMPLEMENT"),
        ("Parent ReqID", "TODO: IMPLEMENT"),
        ("Objective", dic_pro["objective"]),
        ("Responsible", dic_raw["responsible"]),
        ("Phase", dic_raw["phase"]),
        ("MICADO Set-up", dic_raw['micadosetup']),
        ("Calibration Unit setup", dic_raw['calunitsetup']),
        ("Template name", template.__name__),
        ("Input parameters Template", stemplatepars),
        # TODO: Do we need Observing Block name? Where to get it from?
        # ("Observing Block name", "TODO: What is this?"),
        ("Observational frequency", dic_raw['frequency']),
        ("Observation duration", str(dic_raw["duration"])),
        ("Prerequisites", dic_raw['prerequisites']),
        ("Recipe frequency", dic_pro["frequency"]),
        ("Input products", sinput),
        ("Output products", f"{dataitem.name_eso()}_"),
        ("QC parameters", sqc),
        ("DRL Recipe ID", dataitem.recipe.prop_type.name_eso()),
        ("Procedure", dic_pro['procedure']),
        ("Mathematical description", dic_pro['math']),
        ("Remarks", dic_pro["remarks"]),
    ])

    dic2 = {k: str(v).splitlines() for k, v in dic1.items()}

    # Get maximum width of keys and values.
    lenk = max([len(dd) for dd in dic2.keys()])
    lenv = max([max([len(cc) for cc in dd + [""]]) for dd in dic2.values()])

    shline = "+-" + lenk * "-" + "-+-" + lenv * "-" + "-+"

    lstable = [shline]
    for kk, lines in dic2.items():
        sline = f"| {kk:{lenk}} | {lines[0]:{lenv}} |"
        lstable.append(sline)
        for line in lines[1:]:
            sline = f"| {' ':{lenk}} | {line:{lenv}} |"
            lstable.append(sline)

        lstable.append(shline)

    stable = "\n".join(lstable)

    prosepurpose = f"""**Purpose:**
    
{dic_pro['prosepurpose']}
""" if "prosepurpose" in dic_pro else ""
    proseworkflow = f"""**Algorithm and workflow:**

{dic_pro['proseworkflow']}
""" if "proseworkflow" in dic_pro else ""

    return stable, prosepurpose, proseworkflow


def get_keywords_table():
    """
    Create a table for the keywords.
    """
    columns = OrderedDict([
        (n, [k[n] for k in esokeywords.values()])
        for n in ['name', 'type', 'unit', 'description']
    ])
    sizes = OrderedDict([
        (n, max([len(s) for s in v]))
        for n, v in columns.items()
    ])
    shline = " ".join(["=" * size for size in sizes.values()])
    sheader = " ".join([f"{n:{s}}" for n, s in sizes.items()])
    sskeyword = [
        " ".join([f"{(v[k].replace(' ', '.') if k == 'name' else v[k]):{s}}" for k, s in sizes.items()])
        for v in esokeywords.values()
        if '____' not in v['name']
    ]
    stable = "\n".join([shline, sheader, shline] + sskeyword + [shline])
    return stable


def main():
    """
    Convert all dataitems.
    """
    dataitems_img = [MasterDark, Detrend, IllumCorr, MasterFlat]

    for dataitem in dataitems_img:
        stable, prosepurpose, proseworkflow = table_from_dataitem(dataitem)
        # TODO: Fully settle on ESO names in the documentation, also for filenames?
        fname = dataitem.__name__
        sname = dataitem.name_eso()
        sline = "^" * len(sname)
        sfigure = f"""
.. figure:: generated/{fname}.png

    Workflow diagram for {sname}
"""
        srst = f""".. _{sname.lower()}:

{sname}
{sline}

{prosepurpose}
{proseworkflow}

{stable}

{sfigure}
"""

        fn_rst = PATH_DOCS_GENERATED / f"{fname}.rst"
        with open(fn_rst, 'w') as f:
            f.write(srst)

        recipestep: RecipeStep = dataitem.recipe.prop_type.process.prop_type
        while recipestep is not None:
            srecipestep = f"""{recipestep.name_eso()}
{'^' * len(recipestep.name_eso())}

{recipestep.__doc__}

.. figure:: generated/{recipestep.__name__}.png
"""

            fn_rst = PATH_DOCS_GENERATED / f"{recipestep.__name__}.rst"
            with open(fn_rst, 'w') as f:
                f.write(srecipestep)

            recipestep = recipestep.get_previous_step()

    skeywords = get_keywords_table()
    fn_rst_keywords = PATH_DOCS_GENERATED / "keywordstable.rst"
    with open(fn_rst_keywords, 'w') as f:
        f.write(skeywords)


if __name__ == '__main__':
    main()
