# -*- coding: utf-8 -*-
"""
Creating YAML files from ESO DID files.
"""

from collections import OrderedDict
import json
import glob
import os
from pathlib import Path
import yaml


# TODO find workaround for this hack. This is used to find the path
#  in did_test.py during the testing in conda build
PATH_DID = Path(__file__).resolve().parent.parent.parent / "datainterface" / "did"


def parse_param(lines):
    """
    Parse set of lines.

Parameter Name:    PRO RECi RAWi NAME
Class:             header
Context:           process
Type:              string
Value Format:      %45s
Unit:
Comment Field:     File name of raw frame
Description:       File name of raw frame.

The field Comment Format may also be named Comment Field.

TODO: Use a proper parser
TODO: Fix multiline descriptions
    """
    attr_names = OrderedDict((
        ('name', ('Parameter Name',)),
        ('class', ('Class',)),
        ('context', ('Context',)),
        ('type', ('Type',)),
        ('value', ('Value Format',)),
        ('unit', ('Unit',)),
        ('comment', ('Comment Format', 'Comment Field', 'Comment field')),
        ('default', ('Default', )),
        ('description', ('Description',)),
    ))

    # TODO: Fix this dirty hack for optional attributes.
    if lines[7][1:7] != 'efault':
        lines.insert(7, "Default: None")

    assert len(lines) >= len(attr_names)
    for attrnt, line in zip(attr_names.values(), lines):
        assert any(line.startswith(attrn) for attrn in attrnt)

    dict1 = OrderedDict([
        (an, line.split(":", maxsplit=1)[1].strip())
        for an, line in zip(attr_names.keys(), lines)
    ])

    if dict1['name'] == '"        "  (8 ASCII blanks)':
        # See ESO-VLT-DIC.PRIMARY
        # TODO add test
        dict1['name'] = "________"

    desc_extra = " ".join(l.strip() for l in lines[len(attr_names):])
    if desc_extra:
        dict1['description'] += " " + desc_extra

    return dict(dict1)


def parse_did(fn):
    """
    Parse a DID into a dictionary.
    """
    try:
        lines0 = open(fn).readlines()
    except UnicodeDecodeError:
        # e.g. ESO-VLT-DIC.MUSE_CFG
        lines0 = open(fn, encoding="windows-1252").readlines()

    lines1 = [l.strip() for l in lines0]
    lines2 = [l for l in lines1 if len(l) and not l.startswith("#")]
    paramgroups = []
    paramsstarted = False
    paramlines = []
    while lines2:
        line = lines2.pop(0)
        if line.startswith("Parameter Name:"):
            if not paramsstarted:
                paramsstarted = True
            else:
                paramgroups.append(paramlines)
                paramlines = []
            paramlines.append(line)
        elif paramsstarted:
            paramlines.append(line)

    # Add the last parameter.
    paramgroups.append(paramlines)
    params = [parse_param(lines) for lines in paramgroups]
    return params


def make_hierarchy(params):
    """
    Convert parameters to a hierarchy.
    """
    lps = []
    dicti = {
        'subsystems': {},
        'parameters': {},
    }
    for p in params:
        subsystems = p["name"].split()[:-1]
        paramname = p["name"].split()[-1]
        p1 = " ".join(p["name"].split())
        p2 = " ".join(subsystems)
        if p1 in lps:
            print(f"3 {p1} is duplicate")
            continue
        elif any(l.startswith(p1) for l in lps):
            print(f"1 {p1} used in lps")
            continue
        if p2 in lps:
            print(f"2 {p1} base in lps")
            continue
        lps.append(p1)
        dicta = dicti
        for k in subsystems:
            if k not in dicta['subsystems']:
                dicta['subsystems'][k] = {
                    'parameters': {},
                    'subsystems': {},
                }
            dicta = dicta['subsystems'][k]
        dicta['parameters'][paramname] = p
    return dicti


def main(path_did=None, path_yaml=None):
    """
    Convert VODML/XML to Python classes.
    """
    if path_did is None:
        path_did = PATH_DID
    fns = glob.glob(str(path_did / "ESO-*-DIC.*"))
    print(fns)
    if path_yaml is None:
        path_yaml = path_did.parent / "yaml"
    assert path_yaml.exists()
    for fn in fns:
        print()
        print("============")
        print(fn)
        params = parse_did(fn)
        fn_json = path_yaml / (os.path.basename(fn) + ".flat.json")
        with open(fn_json, 'w') as f:
            json.dump(params, f, indent=4)
        fn_yaml = path_yaml / (os.path.basename(fn) + ".flat.yaml")
        with open(fn_yaml, 'w') as f:
            yaml.dump(params, f, indent=4)
        params_hierarch = make_hierarchy(params)
        fn_json = path_yaml / (os.path.basename(fn) + ".hierarch.json")
        with open(fn_json, 'w') as f:
            json.dump(params_hierarch, f, indent=4)
        fn_yamlh = path_yaml / (os.path.basename(fn) + ".hierarch.yaml")
        with open(fn_yamlh, 'w') as f:
            yaml.dump(params_hierarch, f, indent=4)


if __name__ == '__main__':
    main()
